"""Constants used in package

Attributes:
  PKG_INDICES (dict): Key is language abbreviation, such as ``en``, ``es``. Value is the name of pre-trained package 
"""

PKG_INDICES = {
  "en": "en_core_web_trf",
  "es": "es_dep_news_trf",
  "de": "de_dep_news_trf",
  "fr": "fr_dep_news_trf",
}




def merge_range(first, second):
  """Merge two tuples, which have two elements respectively 

  Args:
    first (tuple): (start, end)
    second (tuple): (start, end)

  Returns:
    list: merged tuples


  Examples:
    >>> merge_range((1, 3), (4, 8))
    [(1, 3), (4, 8)]

    >>> merge_range((1, 3), (3, 8))
    [(1, 3), (3, 8)]

    >>> merge_range((1, 3), (2, 8))
    [(1, 8)]
  """

  if first[1] <= second[0] or first[0] >= second[1]:
    return [first, second]
  else:
    return [(min(first[0], second[0]), max(first[1], second[1]))]


def merge(ranges):
  """Merge a list of tuples, which have two elements respectively 

  Args:
    ranges (list): [(start1, end1), (start2, end2), ...]

  Returns:
    list: merged tuples


  Examples:
    >>> merge([(1, 4),(3, 6), (4, 7), (9, 12)])
    [(1, 7), (9, 12)]
  """

  if len(ranges) == 0:
    return []
  ordered = sorted(ranges, key=lambda pair: pair[0])
  result = [ordered[0]] 
  for e in ordered[1:]:
    merged = merge_range(result[-1], e)
    del result[-1]
    result.extend(merged)
  return result






# -*- coding: utf-8 -*-
from spacy.tokens import Doc, Span
from spacy.language import Language 
from importlib import import_module


class NounPhraseRecognizer:
  """Customerized component to detect noun phrases in ``spacy.tokens.Doc`` object. The ``NP`` values are stored in ``doc._.noun_phrases``,

  Attributes:
    ext_name (str): customized extension field name
  """

  def __init__(self, nlp: Language):
    """Initialize the pipeline component. The shared nlp instance is used to initialize the matcher.
    
    Args:
      nlp (spacy.Language): language environment
    """
    lang = nlp.meta["lang"]
    self.ext_name = "noun_phrases"
    self.rule_module_name = 'phrase_detective.{}.noun_phrases'.format(lang)
    self.nlp = nlp

    if not Doc.has_extension(self.ext_name):
      Doc.set_extension(self.ext_name, default=[])

  def __call__(self, doc: Doc):
    """Apply the pipeline component on a ``Doc`` object and modify it if matches are found.

    Returns:
      Doc: with customized extension ``doc._.noun_phrases``
    """

    rule_module = import_module(self.rule_module_name) 
    fn = getattr(rule_module, 'search_out')
    doc._.noun_phrases = fn(doc, self.nlp) 
    return doc




"""Define regular expression usage of ``spacy.matcher.Matcher`` for ``German``
"""

# VERB Knowledge pattern
DE_VERB = [
  [{"POS": "VERB"}],
  [{"TAG": "VAFIN"}],
  [{"TAG": "VAIMP"}],
  [{"TAG": "VAINF"}],
  [{"TAG": "VAPP"}],
  [{"TAG": "VMFIN"}],
  [{"TAG": "VMINF"}],
  [{"TAG": "VMPP"}],
  [{"TAG": "VVFIN"}],
  [{"TAG": "VMIMP"}],
  [{"TAG": "VMINF"}],
  [{"TAG": "VVIZU"}],
  [{"TAG": "VVPP"}],
]




from spacy.matcher import Matcher, DependencyMatcher
from phrase_detective.lib import merge

def search_out(doc, nlp):
  """Search for prepositions with verb 

  Args:
    doc (spacy.tokens.Doc): doc to be analyzed
    nlp (spacy.language.Language): context language

  Returns:
    list: list of spacy.tokens.Span
  """
  result = []
  
  token_matcher = Matcher(nlp.vocab)
  dep_matcher = DependencyMatcher(nlp.vocab)
  
  token_patterns = [
    [{"POS": "NOUN"}],
    [{"POS": "PRON"}],
    [{"POS": "PROPN"}],
    [{"POS": "PROPN"}, {"POS": "CCONJ", "DEP": "cd"}, {"POS": "PROPN", "DEP": "cj"}],
  ]
  token_matcher.add("token_NP", token_patterns)
  token_matches = token_matcher(doc)

  token_refined_matches = merge([(start, end) for _, start, end in token_matches])


  dep_patterns = [
    [
      {
        "RIGHT_ID": "noun",
        "RIGHT_ATTRS": {"POS": "NOUN"}
      },
      {
        "LEFT_ID": "noun",
        "REL_OP": ">",
        "RIGHT_ID": "noun_mod",
        "RIGHT_ATTRS": {"DEP": {"IN": ["ag", "nk", "pg", "op"]}, "POS": {"NOT_IN": ["PART", "ADV", "PRON", "CCONJ"]}}
      },
    ],
    [
      {
        "RIGHT_ID": "pron",
        "RIGHT_ATTRS": {"POS": "PRON"}
      },
      {
        "LEFT_ID": "pron",
        "REL_OP": ">",
        "RIGHT_ID": "pron_mod",
        "RIGHT_ATTRS": {"DEP": {"IN": ["ag"]}, "POS": {"IN": ["NOUN"]}}
      },
    ],
    [
      {
        "RIGHT_ID": "noun",
        "RIGHT_ATTRS": {"POS": "NOUN"}
      },
      {
        "LEFT_ID": "noun",
        "REL_OP": ">",
        "RIGHT_ID": "noun_mod",
        "RIGHT_ATTRS": {"DEP": "mnr", "LOWER":{"IN": ["an", "nach", "vom", "von", "auf", "zu", "zur", "zwischen", "aus"]}}
      },
    ],
    [
      {
        "RIGHT_ID": "propn",
        "RIGHT_ATTRS": {"POS": "PROPN"}
      },
      {
        "LEFT_ID": "propn",
        "REL_OP": ">",
        "RIGHT_ID": "propn_mod",
        "RIGHT_ATTRS": {"DEP": {"IN": ["pnc", "nk"]}, "POS":{"IN": ["NOUN", "DET", "PROPN"]}}
      },
    ],

  ]

  dep_matcher.add("dep_NP", dep_patterns)
  dep_matches = dep_matcher(doc)
  dep_refined_matches = []
  for _, (noun, desp) in dep_matches:
    desp_tree = [e.i for e in doc[desp].subtree]
    length_valid = len(desp_tree) == max(desp_tree) - min(desp_tree) + 1
    noun_valid = any([noun<=end and noun>=start for start, end in token_refined_matches])
    if length_valid and noun_valid:
      desp_tree.append(noun)
      desp_tree.sort()
      dep_refined_matches.append((min(desp_tree),max(desp_tree)+1))
      
  matches = token_refined_matches + dep_refined_matches
  refined_matches = merge(matches)

  for start, end in refined_matches:
    if end-start > 1 and all([e.pos_!="PUNCT" for e in doc[start:end]]):
      np = doc[start: end]
      result.append(np)

  return result
 
  



"""Define regular expression usage of ``spacy.matcher.Matcher`` for ``English``
"""

# VERB Knowledge pattern
EN_VERB = [
  [{"POS": "VERB"}],
  [{"TAG": "VB"}],
  [{"TAG": "VBZ"}],
  [{"TAG": "VBN"}],
  [{"TAG": "VBP"}],
  [{"TAG": "VBG"}],
  [{"TAG": "VBD"}],
]




# -*- coding: utf-8 -*-
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span, Token
from spacy.language import Language 
from phrase_detective.regx import REGX 

class VerbKnowledgeRecognizer:
  """Customerized component to detect verb knowledge in ``spacy.tokens.Doc`` object. The corresponding values are stored in ``doc._.verbs``.

  Attributes:
    ext_name (str): customized extension field name
    matcher (spacy.mathcer.Matcher): Rule maker for detecting ``VERB``
  """

  def __init__(self, nlp: Language):
    """Initialize the pipeline component. The shared nlp instance is used to initialize the matcher.
    
    Args:
      nlp (spacy.Language): language environment
    """

    self.nlp = nlp
    self.matcher = Matcher(nlp.vocab)
    self.ext_name = "verbs"

    lang = nlp.meta["lang"]
    verb_patterns = REGX[lang] 

    self.matcher.add("VERB", verb_patterns)

    if not Doc.has_extension(self.ext_name):
      Doc.set_extension(self.ext_name, default=[])

  def __call__(self, doc):
    """Apply the pipeline component on a Doc object and modify it if matches are found.

    Returns:
      Doc: with customized extensions `doc._.verbs` 
    """
    matches = self.matcher(doc)

    verbs = []
    ids = list(dict.fromkeys([start for _, start, end in matches if end-start==1]))
    for e in ids:
      v = doc[e]
      verbs.append(v)
    doc._.verbs = verbs 

    return doc



"""Regular Expression used to analyze phrases

.. code:: python

  from phrase_detective.regx import REGX

  print(REGX)
"""

from .en import EN_VERB 
from .es import ES_VERB
from .de import DE_VERB

REGX = {
  "en": EN_VERB,
  "es": ES_VERB,
  "de": DE_VERB,
}



from spacy.matcher import DependencyMatcher, Matcher
from phrase_detective.lib import merge

def search_out(doc, nlp):
  """Search for prepositions with verb 

  Args:
    doc (spacy.tokens.Doc): doc to be analyzed
    nlp (spacy.language.Language): context language

  Returns:
    list: list of spacy.tokens.Span
  """
  result = []

  token_matcher = Matcher(nlp.vocab)
  dep_matcher = DependencyMatcher(nlp.vocab)
  
  token_patterns = [
    [{"LOWER": {"IN": ["más", "mneos"]}}, {"POS": "NOUN"}, {"LOWER": "que"}, {}],
    [{"POS": "NOUN"}, {"POS": "PRON", "DEP": "compound"}],
    [{"POS": "DET", "DEP": {"IN": ["det", "amod"]}, "OP": "*"}, {"POS": "NUM", "DEP": "nummod", "OP": "?"}, {"POS": "ADJ", "DEP": "amod", "OP": "*"}, {"POS": "NOUN"}],
    [{"POS": "DET", "DEP": {"IN": ["det", "amod"]}, "OP": "*"}, {"POS": "ADV", "DEP": "advmod", "OP": "?"}, {"POS": "ADJ", "DEP": "amod", "OP": "+"}, {"POS": "NOUN"}],
  ]
  token_matcher.add("token_NP", token_patterns)
  token_matches = token_matcher(doc)

  token_refined_matches = merge([(start, end) for _, start, end in token_matches])


  dep_patterns = [
    [
      {
        "RIGHT_ID": "noun",
        "RIGHT_ATTRS": {"POS": "NOUN"}
      },
      {
        "LEFT_ID": "noun",
        "REL_OP": ">",
        "RIGHT_ID": "mod",
        "RIGHT_ATTRS": {"DEP": {"IN": ["nmod", "amod"]}}
      },
    ],
  ]

  dep_matcher.add("dep_NP", dep_patterns)
  dep_matches = dep_matcher(doc)
  dep_refined_matches = []
  for _, (noun, desp) in dep_matches:
    desp_tree = [e.i for e in doc[desp].subtree]
    length_valid = len(desp_tree) == max(desp_tree) - min(desp_tree) + 1
    noun_valid = any([noun<=end and noun>=start for start, end in token_refined_matches])
    if length_valid and noun_valid and noun < desp:
      desp_tree.append(noun)
      desp_tree.sort()
      dep_refined_matches.append((min(desp_tree),max(desp_tree)+1))
      
  matches = token_refined_matches + dep_refined_matches
  refined_matches = merge(matches)

  for start, end in refined_matches:
    if end-start > 1 and all([e.pos_!="PUNCT" for e in doc[start:end]]):
      np = doc[start: end]
      result.append(np)

  return result
   
